#include <iostream>
#include <utility>
#include "seznam.h"

using namespace std;

void vytvor_seznam(Seznam& s)
{
	s.hlava = s.zarazka = new Prvek{ NULA, nullptr };
	s.velikost = 0;
}

void vloz_na_zacatek(Seznam& s, T co)
{
	s.hlava = new Prvek{ co, s.hlava };
	s.velikost++;
}

void vloz_na_konec(Seznam& s, T co)
{
	s.zarazka->data = co;
	s.zarazka->dalsi = new Prvek{ NULA, nullptr };
	s.zarazka = s.zarazka->dalsi;
	s.velikost++;
}

void odstran_prvni(Seznam& s)
{
	if (s.hlava == s.zarazka)
		return; // seznam byl prazdny (je tam pouze zarazka)
	auto pom = s.hlava;
	s.hlava = s.hlava->dalsi;
	delete pom;
	s.velikost--;
}

bool prazdny(Seznam& s)
{
	//return s.hlava == s.zarazka;
	return !s.velikost;
}

size_t velikost_seznamu(Seznam& s)
{
	return s.velikost;
}

void vyprazdni(Seznam& s)
{
	while (!prazdny(s))
	{
		odstran_prvni(s);
	}
}

void zrus(Seznam& s)
{
	vyprazdni(s);
	delete s.hlava; // dealokovat zarazku
	s.hlava = s.zarazka = nullptr; // nastavit udaje v promenne "s"
}


Prvek* najdi(Seznam& s, T hodnota) // elegantnejsi reseni nez 7. 11. 2022
{
	Prvek* pom = s.hlava;
	s.zarazka->data = hodnota; // pomucka, abychom to VZDY nasli
	while (pom->data != hodnota)
		pom = pom->dalsi; // posuneme se dal
	s.zarazka->data = NULA; // nalezli jsme. Zarazku dame do puvodniho stavu
	if (pom == s.zarazka) // nasli jsme to az na fiktivnim konci?
		return nullptr; // nenalezeno
	else
		return pom;
}

void vloz_za_prvek(Seznam& s, T hledane, T pridavane) // lepsi nazev: "vloz_za_hodnotu"
{
	Prvek* nalezeno;
	nalezeno = najdi(s,hledane);
	if (!nalezeno) // "najdi" vratilo nullptr (tj. nenalezeno)
		vloz_na_konec(s, pridavane); // vlozime na konec
	else
		nalezeno->dalsi = new Prvek{ pridavane, nalezeno->dalsi }; // vlozime nekam "doprostred"
	s.velikost++; // aktualizujeme pocet prvku
}

void odstran_za(Seznam& s, Prvek* za_kterym)
{
	if (za_kterym->dalsi == s.zarazka)
		return; // neodstranujeme nic ("za_kterym" je posledni prvek, za nim je jenom zarazka)
	auto odstranovany = za_kterym->dalsi;
	za_kterym->dalsi = odstranovany->dalsi;
	delete odstranovany;
	s.velikost--;
}

void odstran_hodnotu(Seznam& s, T hodnota) // elegantnejsi verze - s vyuzitim funkce "odstran_za"
{
	Prvek* nalezen = najdi(s, hodnota);
	if (!nalezen)
		return; // hodnota v seznamu neni, nemusime nic odstranovat
	Prvek* za_mnou = nalezen->dalsi;
	nalezen->data = za_mnou->data; // kopie dat
	// nalezen->dalsi nebudeme menit, protoze to udela funkce "odstran_za"
	if (za_mnou == s.zarazka) // co kdyz "nalezen" byl posledni prvek? Musime nastavit novou zarazku!
	{
		s.zarazka = nalezen;
		delete za_mnou; // stara zarazka uz neni potreba
		s.velikost--;
	}
	else // kdyz "nalezen" nebyl poslednim prvkem
		odstran_za(s, nalezen); // provede dealokaci prvku a nastaveni velikosti seznamu
}

Prvek* minimum(Seznam& s)
{
	Prvek *min_prvek, *pom; // dva pomocne ukazatele
	if (prazdny(s))
		return nullptr; // prazdny seznam nema minimum
	min_prvek = pom = s.hlava; // zaciname 1. prvkem
	while (pom != s.zarazka) // dokud nejsme na konci
	{
		if (pom->data < min_prvek->data) // nasli jsme mensi hodnotu
			min_prvek = pom; // novy prvek, kde je minimum
		pom = pom->dalsi; // posuneme se dal
	}
	return min_prvek;
}

void vypis(Seznam& s)
{
	auto pom = s.hlava;
	while (pom != s.zarazka)
	{
		cout << pom->data << " "; // mezi hodnotami vypiseme mezeru
		pom = pom->dalsi;
	}
	cout << endl; // nakonec odradkujeme
}

void serad(Seznam& s)
{
	Seznam pomocny; // pomocny seznam, kam dame serazene prvky
	vytvor_seznam(pomocny);
	while (!prazdny(s))
	{
		Prvek* nejmensi = minimum(s);
		vloz_na_konec(pomocny, nejmensi->data);
		odstran_hodnotu(s, nejmensi->data);
	}
	swap(s.hlava, pomocny.hlava); // zamena promennych (funkce "swap" je z <utility>)
	swap(s.zarazka, pomocny.zarazka);
	swap(s.velikost, pomocny.velikost);
	zrus(pomocny);	
}
