#include <iostream>
#include <utility>
#include "seznam.h"

using namespace std;

void vytvor_seznam(Seznam& s)
{
	s.hlava = s.zarazka = new Prvek{ NULA, nullptr };
	// s.velikost = 0;
}

void vloz_na_zacatek(Seznam& s, T co)
{
	s.hlava = new Prvek{ co, s.hlava };
	// s.velikost++;
}

void vloz_na_konec(Seznam& s, T co)
{
	s.zarazka->data = co;
	s.zarazka->dalsi = new Prvek{ NULA, nullptr };
	s.zarazka = s.zarazka->dalsi;
	// s.velikost++;
}

void odstran_prvni(Seznam& s)
{
	if (prazdny(s))
		return;
	auto pom = s.hlava;
	s.hlava = s.hlava->dalsi;
	delete pom;
	// s.velikost--;
}

bool prazdny(Seznam& s)
{
	return s.hlava == s.zarazka;
	// return !s.velikost;
}

/* size_t velikost_seznamu(Seznam& s)
{
	return s.velikost;
} */

void vyprazdni(Seznam& s)
{
	while (!prazdny(s))
	{
		odstran_prvni(s);
	}
}

void zrus(Seznam& s)
{
	vyprazdni(s);
	delete s.hlava; // dealokovat zarazku
	s.hlava = s.zarazka = nullptr; // nastavit udaje v promenne "s"
}


Iterator_seznamu najdi(Seznam& s, T hodnota) // pomoci iteratoru seznamu
{
	auto p = zacatek(s);
	auto k = konec(s);
	while (data(p) != hodnota && nerovna_se(p,k))
	{
		dalsi(p);
	}
	return p; // nenalezeno => konec (tj. zarazka)
}

void vloz_za(T pridavane, Iterator_seznamu kam) // s vyuzitim iteratoru seznamu
{
	kam.aktualni_prvek->dalsi = new Prvek{ pridavane, kam.aktualni_prvek->dalsi }; // vlozime nekam "doprostred"
	// s.velikost++; // aktualizujeme pocet prvku
}

void odstran_za(Iterator_seznamu iter)
{
	Prvek* za_kterym = iter.aktualni_prvek;
	Prvek* odstranovany = za_kterym->dalsi;

	if (odstranovany == iter.s->zarazka)
		return; // neodstranujeme nic ("za_kterym" je posledni prvek, za nim je jenom zarazka)
	za_kterym->dalsi = odstranovany->dalsi;
	delete odstranovany;
	// s.velikost--;
}

void odstran(Iterator_seznamu iter) // s vyuzitim iteratoru seznamu
{
	Prvek* aktualni = iter.aktualni_prvek;
	Seznam& s = *iter.s;
	if (aktualni->dalsi == s.zarazka) // co kdyz iterator je na poslednim prvku? Musime nastavit novou zarazku!
	{
		s.zarazka = aktualni;
		delete aktualni->dalsi; // stara zarazka uz neni potreba
		aktualni->dalsi = nullptr;
		// s.velikost--;
	}
	else // kdyz "nalezen" nebyl poslednim prvkem
	{
		aktualni->data = aktualni->dalsi->data; // kopie hodnoty
		odstran_za(iter); // provede dealokaci prvku vpravo
	}
}

Iterator_seznamu minimum(Seznam& s)
{
	if (prazdny(s))
		return konec(s); // prazdny seznam nema minimum
	Iterator_seznamu min_prvek = zacatek(s); // zaciname 1. prvkem seznamu
	Iterator_seznamu pom; // inicializujeme to ve FOR
		
	for (pom = zacatek(s); nerovna_se(pom, konec(s)); dalsi(pom)) // dokud nejsme na konci
	{
		if (data(pom) < data(min_prvek)) // nasli jsme mensi hodnotu
			min_prvek = pom; // novy prvek, kde je minimum
	}
	return min_prvek;
}

void vypis(Seznam& s)
{
	for (Iterator_seznamu p = zacatek(s); nerovna_se(p, konec(s)); dalsi(p))
		{
			cout << data(p) << endl;
		}

	/* bez iteratoru to bylo takto:
	auto pom = s.hlava;
	while (pom != s.zarazka)
	{
		cout << pom->data << endl; 
		pom = pom->dalsi;
	} */
}

void serad(Seznam& s)
{
	Seznam pomocny; // pomocny seznam, kam dame serazene prvky
	vytvor_seznam(pomocny);
	while (!prazdny(s))
	{
		auto nejmensi = minimum(s); // iterator seznamu
		vloz_na_konec(pomocny, data(nejmensi));
		odstran(nejmensi);
	}
	prohod(s, pomocny);
	zrus(pomocny);	
}

void prohod(Seznam& s1, Seznam& s2)
{
	swap(s1.hlava, s2.hlava);
	swap(s1.zarazka, s2.zarazka);
}

// Iterator ukazujici na hlavu/zarazku

Iterator_seznamu zacatek(Seznam& s)
{
	return Iterator_seznamu{ s.hlava, &s };
}

Iterator_seznamu konec(Seznam& s)
{
	return Iterator_seznamu{ s.zarazka, &s };
}

// Operace s iteratorem seznamu

void dalsi(Iterator_seznamu& ten)
{
	if (ten.aktualni_prvek != ten.s->zarazka)
	{
		ten.aktualni_prvek = ten.aktualni_prvek->dalsi;
	}
}

T data(Iterator_seznamu kde)
{
	return kde.aktualni_prvek->data;
}

void vloz_hodnotu(Iterator_seznamu kam, T co)
{
	kam.aktualni_prvek->data = co;
}

bool rovna_se(Iterator_seznamu a, Iterator_seznamu b)
{
	return (a.s == b.s) && (a.aktualni_prvek == b.aktualni_prvek);
}

bool nerovna_se(Iterator_seznamu a, Iterator_seznamu b)
{
	return !rovna_se(a, b);
}
